@@ -1,27 +1,25 @@
+from __future__ import annotations
 import asyncio
 import signal
-import sys
-import time
-from enum import Enum
-from logging import Logger
 from typing import Any, Dict, List, Optional, Union
 
 import backoff
 import httpx
-import loguru
 import typer
+import loguru
 
 from devtools import pformat
 from pydantic import BaseModel, Field, parse_obj_as
 
 from servo import api
+from servo.api import descriptor_to_adjustments
 from servo.assembly import Assembly, BaseServoConfiguration
 from servo.configuration import Optimizer
 from servo.errors import ConnectorError
 from servo.logging import ProgressHandler
 from servo.servo import Events, Servo
-from servo.types import Control, Description, Measurement
-from servo.utilities import commandify
+from servo.types import Adjustment, Control, Duration, Description, Measurement
+from servo.utilities import commandify, value_for_key_path
 
 
 class Runner(api.Mixin):
@@ -40,11 +38,11 @@ class Runner(api.Mixin):
         return self.assembly.servo
 
     @property
-    def configuration(self) -> BaseServoConfiguration:
-        return self.servo.configuration
+    def config(self) -> BaseServoConfiguration:
+        return self.servo.config
 
     @property
-    def logger(self) -> Logger:
+    def logger(self) -> loguru.Logger:
         return self.servo.logger
     
     def display_banner(self) -> None:
@@ -57,8 +55,6 @@ class Runner(api.Mixin):
         )
         typer.secho(banner, fg=typer.colors.BRIGHT_BLUE, bold=True)
                 
-        name_st = typer.style("name", fg=typer.colors.CYAN, bold=False)
-        version_st = typer.style("version", fg=typer.colors.WHITE, bold=True)
         types = Assembly.all_connector_types()
         types.remove(Servo)
         
@@ -68,7 +64,7 @@ class Runner(api.Mixin):
             version = typer.style(str(c.version), fg=typer.colors.WHITE, bold=True)
             names.append(f"{name}-{version}")
         version = typer.style(f"v{Servo.version}", fg=typer.colors.WHITE, bold=True)
-        codename = typer.style("the awakening", fg=typer.colors.MAGENTA, bold=False)
+        codename = typer.style("woke up like this", fg=typer.colors.MAGENTA, bold=False)
         initialized = typer.style("initialized", fg=typer.colors.BRIGHT_GREEN, bold=True)        
         
         typer.secho(f"{version} \"{codename}\" {initialized}")
@@ -109,34 +105,17 @@ class Runner(api.Mixin):
 
         return aggregate_measurement
 
-    async def adjust(self, param) -> dict:
-        self.logger.info("Adjusting...")
-        self.logger.trace(pformat(param))
-
-        results: List[EventResult] = await self.servo.dispatch_event(Events.ADJUST, param)
-        for result in results:
-            # TODO: Should be modeled
-            adjustment = result.value
-            status = adjustment.get("status", "undefined")
-
-            if status == "ok":
-                self.logger.info(f"{result.connector.name} - Adjustment completed")
-                return adjustment
-            else:
-                raise ConnectorError(
-                    'Adjustment driver failed with status "{}" and message:\n{}'.format(
-                        status, str(adjustment.get("message", "undefined"))
-                    ),
-                    status=status,
-                    reason=adjustment.get("reason", "undefined"),
-                )
-
-        # TODO: Model a response class
-        return {}
+    async def adjust(self, adjustments: List[Adjustment], control: Control) -> None:
+        summary = f"[{', '.join(list(map(str, adjustments)))}]"
+        self.logger.info(f"Adjusting... {summary}")
+        self.logger.trace(pformat(adjustments))
+        
+        await self.servo.dispatch_event(Events.ADJUST, adjustments)
+        self.logger.info(f"Adjustment completed {summary}")
 
     async def exec_command(self):
         cmd_response = await self._post_event(api.Event.WHATS_NEXT, None)
-        self.logger.debug(f"What's Next? => {cmd_response.command}")
+        self.logger.info(f"What's Next? => {cmd_response.command}")
         self.logger.trace(pformat(cmd_response))
 
         try:
@@ -159,24 +138,15 @@ class Runner(api.Mixin):
                 await self._post_event(api.Event.MEASUREMENT, param)
 
             elif cmd_response.command == api.Command.ADJUST:
-                # # TODO: This needs to be modeled
-                # oc"{'cmd': 'ADJUST', 'param': {'state': {'application': {'components': {'web': {'settings': {'cpu': {'value': 0.225}, 'mem': {'value': 0.1}}}}}}, 'control': {}}}"
+                adjustments = descriptor_to_adjustments(cmd_response.param["state"])
+                control = Control(**cmd_response.param.get("control", {}))
+                await self.adjust(adjustments, control)
 
-                # TODO: Why do we do this nonsense??
-                # create a new dict based on p['state'] (with its top level key
-                # 'application') which also includes a top-level 'control' key, and
-                # pass this to adjust()
-                new_dict = cmd_response.param["state"].copy()
-                new_dict["control"] = cmd_response.param.get("control", {})
-                adjustment = await self.adjust(new_dict)
+                # TODO: Model a response class ("Adjusted"?) -- map errors to the response
+                # If no errors, report state matching the request
+                reply = { "status": "ok", "state": cmd_response.param["state"] }
 
-                # TODO: What works like this and why?
-                if (
-                    "state" not in adjustment
-                ):  # if driver didn't return state, assume it is what was requested
-                    adjustment["state"] = cmd_response.param["state"]
-
-                components_dict = adjustment["state"]["application"]["components"]
+                components_dict = cmd_response.param["state"]["application"]["components"]
                 components_count = len(components_dict)
                 settings_count = sum(
                     len(components_dict[component]["settings"])
@@ -186,29 +156,36 @@ class Runner(api.Mixin):
                     f"Adjusted: {components_count} components, {settings_count} settings"
                 )
 
-                await self._post_event(api.Event.ADJUSTMENT, adjustment)
+                await self._post_event(api.Event.ADJUSTMENT, reply)
 
             elif cmd_response.command == api.Command.SLEEP:
                     # TODO: Model this
-                    duration = int(cmd_response.param.get("duration", 120))
-                    self.logger.info(f"Sleeping {duration} sec.")
-                    await asyncio.sleep(duration)
+                    duration = Duration(cmd_response.param.get("duration", 120))
+                    status = value_for_key_path(cmd_response.param, "data.status", None)
+                    reason = value_for_key_path(cmd_response.param, "data.reason", "unknown reason")
+                    msg = f"{status}: {reason}" if status else f"{reason}"
+                    self.logger.info(f"Sleeping for {duration} ({msg}).")
+                    await asyncio.sleep(duration.total_seconds())
 
             else:
                 raise ValueError(f"Unknown command '{cmd_response.command.value}'")
 
-        except Exception as error:
-            self.logger.exception(f"{cmd_response.command} command failed!")
-            param = dict(status="failed", message=_exc_format(error))
+        except Exception as error:            
+            self.logger.exception(f"{cmd_response.command} command failed: {error}")
+            param = dict(status="failed", message=str(error))
             await self.shutdown(asyncio.get_event_loop())
             await self._post_event(cmd_response.command.response_event, param)
 
     async def main(self) -> None:
+        # Setup logging
+        self.progress_handler = ProgressHandler(self.servo.report_progress, self.logger.warning)
+        self.logger.add(self.progress_handler.sink, catch=True)
+
         self.logger.info(
             f"Servo started with {len(self.servo.connectors)} active connectors [{self.optimizer.id} @ {self.optimizer.base_url}]"
         )
-        self.logger.info("Broadcasting startup event...")
-        self.servo.startup()
+        self.logger.info("Dispatching startup event...")
+        await self.servo.startup()
 
         self.logger.info("Saying HELLO.", end=" ")
         await self._post_event(api.Event.HELLO, dict(agent=api.USER_AGENT))
@@ -226,10 +203,8 @@ class Runner(api.Mixin):
         try:
             reason = signal.name if signal else 'shutdown'
             await self._post_event(api.Event.GOODBYE, dict(reason=reason))
-        except Exception as e:
-            self.logger.exception(
-                f"Exception occurred during GOODBYE request: {e}"
-            )
+        except Exception:
+            self.logger.exception(f"Exception occurred during GOODBYE request")
         
         self.logger.info("Dispatching shutdown event...")
         await self.servo.shutdown()
@@ -241,26 +216,37 @@ class Runner(api.Mixin):
 
         self.logger.info(f"Cancelling {len(tasks)} outstanding tasks")
         await asyncio.gather(*tasks, return_exceptions=True)
-
+        
+        self.logger.info("Servo shutdown complete.")
+        await self.logger.complete()
+        
         loop.stop()
     
-    def handle_exception(self, loop, context):
+    def handle_exception(self, loop: asyncio.AbstractEventLoop, context: dict) -> None:
+        self.logger.error(f"asyncio exception handler triggered with context: {context}")
+
         # context["message"] will always be there; but context["exception"] may not
-        msg = context.get("exception", context["message"])
-        self.logger.exception(f"Caught exception: {msg}")
-        self.logger.info("Shutting down...")
-        asyncio.create_task(self.shutdown(loop))
+        exception = context.get("exception")
+        if exception:
+            # FIXME: it is not necessary to re-raise this to get the right logging it is logging None atm
+            try:
+                raise exception
+            except:
+                self.logger.exception(f"exception details: {exception}")
+
+        self.logger.critical("Shutting down due to unhandled exception in asyncio task...")
+
+        # try to shutdown cleanly
+        try:
+            asyncio.create_task(self.shutdown(loop))
+        except Exception as exception:
+            self.logger.exception(f"caught exception trying to schedule shutdown: {exception}")
 
     def run(self) -> None:
         self.display_banner()
 
         # Setup async event loop
         loop = asyncio.get_event_loop()
-
-        # Setup logging
-        handler = ProgressHandler(self.servo)
-        loop.create_task(handler.run())
-        loguru.logger.add(handler)        
         
         signals = (signal.SIGHUP, signal.SIGTERM, signal.SIGINT, signal.SIGUSR1)
         for s in signals:
@@ -273,10 +259,4 @@ class Runner(api.Mixin):
             loop.create_task(self.main())
             loop.run_forever()
         finally:
-            loop.close()
-            self.logger.info("Servo shutdown complete.")
-
-def _exc_format(e):
-    if type(e) is Exception:  # if it's just an Exception
-        return str(e)  # print only the message but not the type
-    return "{type(e).__name__}: {e}"
+            loop.close()            
