@@ -3,3 +3,5 @@ from .duration_str import *
 from .pydantic import *
 from .yaml import *
 from .key_paths import *
+from .hashing import *
+from .subprocess import *
